import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertDoesNotThrow;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertTimeoutPreemptively;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.io.File;
import java.time.Duration;
import java.util.Set;

import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;

import edu.princeton.cs.algs4.In;
import edu.princeton.cs.algs4.StdRandom;

public class SolverTests {
    private static BoggleBoard randomBoard;

    @BeforeAll
    static void makeRandomBoard() {
        StdRandom.setSeed(201);
        randomBoard = new BoggleBoard();
    }

    @Test
    void onlyValidLocationsTest() {
        assertTimeoutPreemptively(Duration.ofSeconds(2), () -> {
            String[] dict = { "PINS", "PINES" };
            BoggleSolver solver = new BoggleSolver(dict);
            BoggleBoard board = new BoggleBoard("board4x4.txt");
            assertDoesNotThrow(() -> {
                solver.getAllValidWords(board);
            }, "solver throws an exception, make sure you are using board.isValidLocation to avoid index errors");
        }, "test took longer than two seconds, make sure you are using prefixes to avoid dead ends and not printing anything");
    }

    @Test
    void wordLengthTest() {
        assertTimeoutPreemptively(Duration.ofSeconds(2), () -> {
            String[] dict = { "TO", "TOP" };
            BoggleSolver solver = new BoggleSolver(dict);
            BoggleBoard board = new BoggleBoard("board-word-length.txt");
            Set<String> words = solver.getAllValidWords(board);
            assertEquals(1, words.size(), "solver allows 2-letter words or words not in the dictionary");
            assertTrue(words.contains("TOP"), "solver did not find the correct word");
        }, "test took longer than two seconds, make sure you are using prefixes to avoid dead ends and not printing anything");
    }

    @Test
    void findsSomeWordsTest() {
        assertTimeoutPreemptively(Duration.ofSeconds(2), () -> {
            In in = new In(new File("dictionary-yawl.txt"));
            String[] dictionary = in.readAllStrings();
            BoggleSolver solver = new BoggleSolver(dictionary);

            Set<String> words = solver.getAllValidWords(randomBoard);
            assertTrue(words.size() > 0, "solver does not find any words");
        }, "test took longer than two seconds, make sure you are using prefixes to avoid dead ends and not printing anything");
    }

    @Test
    void horizontalTest() {
        assertTimeoutPreemptively(Duration.ofSeconds(2), () -> {
            String[] dict = { "DATA", "TYPE" };
            BoggleSolver solver = new BoggleSolver(dict);
            BoggleBoard board = new BoggleBoard("board-horizontal.txt");
            Set<String> words = solver.getAllValidWords(board);
            assertTrue(words.contains("DATA"), "solver did not find horizontal word left to right across top row");
            assertTrue(words.contains("TYPE"), "solver did not find horizontal word right to left across third row");
            assertEquals(2, words.size(), "solver found words not in the dictionary or not on the board");
        }, "test took longer than two seconds, make sure you are using prefixes to avoid dead ends and not printing anything");
    }

    @Test
    void verticalTest() {
        assertTimeoutPreemptively(Duration.ofSeconds(2), () -> {
            String[] dict = { "TREE", "NODE" };
            BoggleSolver solver = new BoggleSolver(dict);
            BoggleBoard board = new BoggleBoard("board-vertical.txt");
            Set<String> words = solver.getAllValidWords(board);
            assertTrue(words.contains("TREE"), "solver did not find vertical word top to bottom along leftmost column");
            assertTrue(words.contains("NODE"),
                    "solver did not find vertical word bottom to top along rightmost column row");
            assertEquals(2, words.size(), "solver found words not in the dictionary or not on the board");
        }, "test took longer than two seconds, make sure you are using prefixes to avoid dead ends and not printing anything");
    }

    @Test
    void diagonalTest() {
        assertTimeoutPreemptively(Duration.ofSeconds(2), () -> {
            String[] dict = { "THEN", "HEN", "THE" };
            BoggleSolver solver = new BoggleSolver(dict);
            BoggleBoard board = new BoggleBoard("board-diagonal.txt");
            Set<String> words = solver.getAllValidWords(board);
            assertTrue(words.contains("THE"), "solver did not find diagonal word (3, 3) to (1, 1)");
            assertTrue(words.contains("HEN"), "solver did not find diagonal word (2, 2) to (0, 0)");
            assertTrue(words.contains("THEN"), "solver did not find diagonal word (3, 3) to (0, 0)");
            assertEquals(3, words.size(), "solver found words not in the dictionary or not on the board");
        }, "test took longer than two seconds, make sure you are using prefixes to avoid dead ends and not printing anything");
    }

    @Test
    void noRepeatsTest() {
        assertTimeoutPreemptively(Duration.ofSeconds(2), () -> {
            String[] dict = { "TEPEE" };
            BoggleSolver solver = new BoggleSolver(dict);
            BoggleBoard board = new BoggleBoard("board4x4.txt");
            Set<String> words = solver.getAllValidWords(board);
            assertFalse(words.contains("TEPEE"), "solver allows letters to be used more than once\n"
                    + "make sure you are using board.visit and board.isVisited");
        }, "test took longer than two seconds, make sure you are using prefixes to avoid dead ends and not printing anything");
    }

    @Test
    void onlyAdjacentTest() {
        assertTimeoutPreemptively(Duration.ofSeconds(2), () -> {
            String[] dict = { "PINT", "DATES" };
            BoggleSolver solver = new BoggleSolver(dict);
            BoggleBoard board = new BoggleBoard("board4x4.txt");
            Set<String> words = solver.getAllValidWords(board);
            assertFalse(words.contains("PINT"), "solver allows non-adjacent letters to form words");
            assertFalse(words.contains("DATES"), "solver allows non-adjacent letters to form words");
        }, "test took longer than two seconds, make sure you are using prefixes to avoid dead ends and not printing anything");
    }

    @Test
    void findsAllWordsTest() {
        assertTimeoutPreemptively(Duration.ofSeconds(2), () -> {
            String[] expectedWords = { "GONG", "GONK", "INGLOBE", "GHI", "BOLT", "KHI", "THING", "UNITE", "HINGE",
                    "THINK", "GETGO", "BOH", "HOB", "LOB", "HOE", "HOH", "BETING", "UNIT", "IKON", "OUK", "LOBE", "EIK",
                    "NOG", "EIKON", "UNGET", "KIN", "HOEING", "OHING", "KING", "BETHINK", "TEGG", "GET", "KIT", "BEIN",
                    "UNBE", "HETING", "KINO", "THIN", "GLOBE", "NOGG", "INK", "NITE", "TING", "GNU", "THE", "UNI",
                    "EHING", "OBE", "TINGE", "TINK", "UNBOLT", "GLOB", "HOHING", "THO", "HING", "BEING", "GOUK", "EGG",
                    "HIE", "GON", "BETH", "NIE", "KON", "EGO", "KINGBOLT", "KITE", "HIN", "KNIT", "HOLT", "KHET",
                    "INGO", "KITHE", "BET", "KITH", "TIE", "HIT", "HET", "TEG", "THEIN", "NIT", "ETH", "TIN" };
            In in = new In(new File("dictionary-yawl.txt"));
            String[] dictionary = in.readAllStrings();
            BoggleSolver solver = new BoggleSolver(dictionary);

            Set<String> words = solver.getAllValidWords(randomBoard);
            assertEquals(expectedWords.length, words.size(),
                    expectedWords.length > words.size() ? "solver does not find all the words"
                            : "solver finds too many words");
            assertArrayEquals(expectedWords, words.toArray(),
                    "solver finds the right number of words, but they aren't the exact right ones");
        }, "test took longer than two seconds, make sure you are using prefixes to avoid dead ends and not printing anything");
    }

    @Test
    void onlyRealWordsTest() {
        assertTimeoutPreemptively(Duration.ofSeconds(2), () -> {
            In in = new In(new File("dictionary-yawl.txt"));
            String[] dictionary = in.readAllStrings();
            BoggleSolver solver = new BoggleSolver(dictionary);
            BoggleBoard board = new BoggleBoard("board4x4.txt");
            Set<String> words = solver.getAllValidWords(board);
            assertFalse(words.contains("SID"), "solver finds words not in the dictionary");
        }, "test took longer than two seconds, make sure you are using prefixes to avoid dead ends and not printing anything");
    }

    @Test
    void differentWordsCanOverlapTest() {
        assertTimeoutPreemptively(Duration.ofSeconds(2), () -> {
            In in = new In(new File("dictionary-yawl.txt"));
            String[] dictionary = in.readAllStrings();
            BoggleSolver solver = new BoggleSolver(dictionary);
            BoggleBoard board = new BoggleBoard("board4x4.txt");
            Set<String> words = solver.getAllValidWords(board);
            assertTrue(words.contains("APE") && words.contains("APT"),
                    "make sure every visited location eventually gets unvisited, as it appears letters are not being reused in different words");
        }, "test took longer than two seconds, make sure you are using prefixes to avoid dead ends and not printing anything");
    }
}
